@@ -24,6 +24,7 @@ Provides a common interface to all backends and certain sevices
 intended to be used by the backends themselves.
 """
 
+import errno
 import os
 import sys
 import socket
@@ -31,20 +32,23 @@ import time
 import re
 import getpass
 import gettext
+import re
+import types
 import urllib
-import urlparse_2_5 as urlparser
+import urlparse
 
 from duplicity import dup_temp
-from duplicity import dup_threading
 from duplicity import file_naming
 from duplicity import globals
 from duplicity import log
+from duplicity import path
 from duplicity import progress
 from duplicity import util
 
 from duplicity.util import exception_traceback
 
-from duplicity.errors import BackendException, FatalBackendError
+from duplicity.errors import BackendException
+from duplicity.errors import FatalBackendException
 from duplicity.errors import TemporaryLoadException
 from duplicity.errors import ConflictingScheme
 from duplicity.errors import InvalidBackendURL
@@ -56,8 +60,22 @@ import duplicity.backends
 # todo: this should really NOT be done here
 socket.setdefaulttimeout(globals.timeout)
 
-_forced_backend = None
 _backends = {}
+_backend_prefixes = {}
+
+# These URL schemes have a backend with a notion of an RFC "network location".
+# The 'file' and 's3+http' schemes should not be in this list.
+# 'http' and 'https' are not actually used for duplicity backend urls, but are needed
+# in order to properly support urls returned from some webdav servers. adding them here
+# is a hack. we should instead not stomp on the url parsing module to begin with.
+#
+# This looks similar to urlparse's 'uses_netloc' list, but urlparse doesn't use
+# that list for parsing, only creating urls.  And doesn't include our custom
+# schemes anyway.  So we keep our own here for our own use.
+#
+# NOTE: this is filled by the respective backends during registering
+uses_netloc = []
+
 
 def import_backends():
     """
@@ -76,28 +94,16 @@ def import_backends():
         if fn.endswith("backend.py"):
             fn = fn[:-3]
             imp = "duplicity.backends.%s" % (fn,)
-            # ignore gio as it is explicitly loaded in commandline.parse_cmdline_options()
-            if fn == "giobackend": continue
             try:
                 __import__(imp)
                 res = "Succeeded"
-                level = log.INFO
             except Exception:
                 res = "Failed: " + str(sys.exc_info()[1])
-                level = log.WARNING
-            log.Log(_("Import of %s %s") % (imp, res), level)
+            log.Log(_("Import of %s %s") % (imp, res), log.INFO)
         else:
             continue
 
 
-def force_backend(backend):
-    """
-    Forces the use of a particular backend, regardless of schema
-    """
-    global _forced_backend
-    _forced_backend = backend
-
-
 def register_backend(scheme, backend_factory):
     """
     Register a given backend factory responsible for URL:s with the
@@ -124,6 +130,39 @@ def register_backend(scheme, backend_factory):
     _backends[scheme] = backend_factory
 
 
+def register_backend_prefix(scheme, backend_factory):
+    """
+    Register a given backend factory responsible for URL:s with the
+    given scheme prefix.
+
+    The backend must be a callable which, when called with a URL as
+    the single parameter, returns an object implementing the backend
+    protocol (i.e., a subclass of Backend).
+
+    Typically the callable will be the Backend subclass itself.
+
+    This function is not thread-safe and is intended to be called
+    during module importation or start-up.
+    """
+    global _backend_prefixes
+
+    assert callable(backend_factory), "backend factory must be callable"
+
+    if scheme in _backend_prefixes:
+        raise ConflictingScheme("the prefix %s already has a backend "
+                                "associated with it"
+                                "" % (scheme,))
+
+    _backend_prefixes[scheme] = backend_factory
+
+
+def strip_prefix(url_string, prefix_scheme):
+    """
+    strip the prefix from a string e.g. par2+ftp://... -> ftp://...
+    """
+    return re.sub('(?i)^' + re.escape(prefix_scheme) + '\+', '', url_string)
+
+
 def is_backend_url(url_string):
     """
     @return Whether the given string looks like a backend URL.
@@ -137,9 +176,9 @@ def is_backend_url(url_string):
         return False
 
 
-def get_backend(url_string):
+def get_backend_object(url_string):
     """
-    Instantiate a backend suitable for the given URL, or return None
+    Find the right backend class instance for the given URL, or return None
     if the given string looks like a local path rather than a URL.
 
     Raise InvalidBackendURL if the URL is not a valid URL.
@@ -147,63 +186,45 @@ def get_backend(url_string):
     if not is_backend_url(url_string):
         return None
 
-    pu = ParsedUrl(url_string)
+    global _backends, _backend_prefixes
 
-    # Implicit local path
+    pu = ParsedUrl(url_string)
     assert pu.scheme, "should be a backend url according to is_backend_url"
 
-    global _backends, _forced_backend
+    factory = None
 
-    if _forced_backend:
-        return _forced_backend(pu)
-    elif not pu.scheme in _backends:
-        raise UnsupportedBackendScheme(url_string)
-    else:
-        try:
-            return _backends[pu.scheme](pu)
-        except ImportError:
-            raise BackendException(_("Could not initialize backend: %s") % str(sys.exc_info()[1]))
+    for prefix in _backend_prefixes:
+        if url_string.startswith(prefix + '+'):
+            factory = _backend_prefixes[prefix]
+            pu = ParsedUrl(strip_prefix(url_string, prefix))
+            break
 
-_urlparser_initialized = False
-_urlparser_initialized_lock = dup_threading.threading_module().Lock()
+    if factory is None:
+        if pu.scheme not in _backends:
+            raise UnsupportedBackendScheme(url_string)
+        else:
+            factory = _backends[pu.scheme]
 
-def _ensure_urlparser_initialized():
+    try:
+        return factory(pu)
+    except ImportError:
+        raise BackendException(_("Could not initialize backend: %s") % str(sys.exc_info()[1]))
+
+
+def get_backend(url_string):
     """
-    Ensure that the appropriate clobbering of variables in the
-    urlparser module has been done. In the future, the need for this
-    clobbering to begin with should preferably be eliminated.
+    Instantiate a backend suitable for the given URL, or return None
+    if the given string looks like a local path rather than a URL.
+
+    Raise InvalidBackendURL if the URL is not a valid URL.
     """
-    def init():
-        global _urlparser_initialized
+    if globals.use_gio:
+        url_string = 'gio+' + url_string
+    obj = get_backend_object(url_string)
+    if obj:
+        obj = BackendWrapper(obj)
+    return obj
 
-        if not _urlparser_initialized:
-            # These URL schemes have a backend with a notion of an RFC "network location".
-            # The 'file' and 's3+http' schemes should not be in this list.
-            # 'http' and 'https' are not actually used for duplicity backend urls, but are needed
-            # in order to properly support urls returned from some webdav servers. adding them here
-            # is a hack. we should instead not stomp on the url parsing module to begin with.
-            #
-            # todo: eliminate the need for backend specific hacking here completely.
-            urlparser.uses_netloc = ['ftp',
-                                     'ftps',
-                                     'hsi',
-                                     'rsync',
-                                     's3',
-                                     'u1',
-                                     'scp', 'ssh', 'sftp',
-                                     'webdav', 'webdavs',
-                                     'gdocs',
-                                     'http', 'https',
-                                     'imap', 'imaps',
-                                     'mega']
-
-            # Do not transform or otherwise parse the URL path component.
-            urlparser.uses_query = []
-            urlparser.uses_fragm = []
-
-            _urlparser_initialized = True
-
-    dup_threading.with_lock(_urlparser_initialized_lock, init)
 
 class ParsedUrl:
     """
@@ -217,16 +238,19 @@ class ParsedUrl:
     Raise InvalidBackendURL on invalid URL's
     """
     def __init__(self, url_string):
-        _ensure_urlparser_initialized()
         self.url_string = url_string
 
+        # Python < 2.6.5 still examine urlparse.uses_netlock when parsing urls,
+        # so stuff our custom list in there before we parse.
+        urlparse.uses_netloc = uses_netloc
+
         # While useful in some cases, the fact is that the urlparser makes
         # all the properties in the URL deferred or lazy.  This means that
         # problems don't get detected till called.  We'll try to trap those
         # problems here, so they will be caught early.
 
         try:
-            pu = urlparser.urlparse(url_string)
+            pu = urlparse.urlparse(url_string)
         except Exception:
             raise InvalidBackendURL("Syntax error in: %s" % url_string)
 
@@ -242,6 +266,8 @@ class ParsedUrl:
 
         try:
             self.path = pu.path
+            if self.path:
+                self.path = urllib.unquote(self.path)
         except Exception:
             raise InvalidBackendURL("Syntax error (path) in: %s" % url_string)
 
@@ -272,23 +298,36 @@ class ParsedUrl:
         self.port = None
         try:
             self.port = pu.port
-        except Exception: # not raised in python2.7+, just returns None
+        except Exception:  # not raised in python2.7+, just returns None
             # old style rsync://host::[/]dest, are still valid, though they contain no port
-            if not ( self.scheme in ['rsync'] and re.search('::[^:]*$', self.url_string)):
-                raise InvalidBackendURL("Syntax error (port) in: %s A%s B%s C%s" % (url_string, (self.scheme in ['rsync']), re.search('::[^:]+$', self.netloc), self.netloc ) )
+            if not (self.scheme in ['rsync'] and re.search('::[^:]*$', self.url_string)):
+                raise InvalidBackendURL("Syntax error (port) in: %s A%s B%s C%s" %
+                                        (url_string, (self.scheme in ['rsync']),
+                                         re.search('::[^:]+$', self.netloc), self.netloc))
+
+        # Our URL system uses two slashes more than urlparse's does when using
+        # non-netloc URLs.  And we want to make sure that if urlparse assuming
+        # a netloc where we don't want one, that we correct it.
+        if self.scheme not in uses_netloc:
+            if self.netloc:
+                self.path = '//' + self.netloc + self.path
+                self.netloc = ''
+                self.hostname = None
+            elif not self.path.startswith('//') and self.path.startswith('/'):
+                self.path = '//' + self.path
 
         # This happens for implicit local paths.
         if not self.scheme:
             return
 
         # Our backends do not handle implicit hosts.
-        if self.scheme in urlparser.uses_netloc and not self.hostname:
+        if self.scheme in uses_netloc and not self.hostname:
             raise InvalidBackendURL("Missing hostname in a backend URL which "
                                     "requires an explicit hostname: %s"
                                     "" % (url_string))
 
         # Our backends do not handle implicit relative paths.
-        if self.scheme not in urlparser.uses_netloc and not self.path.startswith('//'):
+        if self.scheme not in uses_netloc and not self.path.startswith('//'):
             raise InvalidBackendURL("missing // - relative paths not supported "
                                     "for scheme %s: %s"
                                     "" % (self.scheme, url_string))
@@ -300,171 +339,79 @@ class ParsedUrl:
 def strip_auth_from_url(parsed_url):
     """Return a URL from a urlparse object without a username or password."""
 
-    # Get a copy of the network location without the username or password.
-    straight_netloc = parsed_url.netloc.split('@')[-1]
-
-    # Replace the full network location with the stripped copy.
-    return parsed_url.geturl().replace(parsed_url.netloc, straight_netloc, 1)
+    clean_url = re.sub('^([^:/]+://)(.*@)?(.*)', r'\1\3', parsed_url.geturl())
+    return clean_url
 
 
-# Decorator for backend operation functions to simplify writing one that
-# retries.  Make sure to add a keyword argument 'raise_errors' to your function
-# and if it is true, raise an exception on an error.  If false, fatal-log it.
-def retry(fn):
-    def iterate(*args):
-        for n in range(1, globals.num_retries):
-            try:
-                kwargs = {"raise_errors" : True}
-                return fn(*args, **kwargs)
-            except Exception, e:
-                log.Warn(_("Attempt %s failed: %s: %s")
-                         % (n, e.__class__.__name__, util.uexc(e)))
-                log.Debug(_("Backtrace of previous error: %s")
-                          % exception_traceback())
-                if isinstance(e, TemporaryLoadException):
-                    time.sleep(30) # wait longer before trying again
-                else:
-                    time.sleep(10) # wait a bit before trying again
-        # Now try one last time, but fatal-log instead of raising errors
-        kwargs = {"raise_errors" : False}
-        return fn(*args, **kwargs)
-    return iterate
+def _get_code_from_exception(backend, operation, e):
+    if isinstance(e, BackendException) and e.code != log.ErrorCode.backend_error:
+        return e.code
+    elif hasattr(backend, '_error_code'):
+        return backend._error_code(operation, e) or log.ErrorCode.backend_error
+    elif hasattr(e, 'errno'):
+        # A few backends return such errors (local, paramiko, etc)
+        if e.errno == errno.EACCES:
+            return log.ErrorCode.backend_permission_denied
+        elif e.errno == errno.ENOENT:
+            return log.ErrorCode.backend_not_found
+        elif e.errno == errno.ENOSPC:
+            return log.ErrorCode.backend_no_space
+    return log.ErrorCode.backend_error
 
-# same as above, a bit dumber and always dies fatally if last trial fails
-# hence no need for the raise_errors var ;), we really catch everything here
-# as we don't know what the underlying code comes up with and we really *do*
-# want to retry globals.num_retries times under all circumstances
-def retry_fatal(fn):
-    def _retry_fatal(self, *args):
-        try:
-            n = 0
-            for n in range(1, globals.num_retries):
+
+def retry(operation, fatal=True):
+    # Decorators with arguments introduce a new level of indirection.  So we
+    # have to return a decorator function (which itself returns a function!)
+    def outer_retry(fn):
+        def inner_retry(self, *args):
+            for n in range(1, globals.num_retries + 1):
                 try:
-                    self.retry_count = n
                     return fn(self, *args)
-                except FatalBackendError, e:
+                except FatalBackendException as e:
                     # die on fatal errors
                     raise e
-                except Exception, e:
+                except Exception as e:
                     # retry on anything else
-                    log.Warn(_("Attempt %s failed. %s: %s")
-                             % (n, e.__class__.__name__, util.uexc(e)))
                     log.Debug(_("Backtrace of previous error: %s")
                               % exception_traceback())
-                    time.sleep(10) # wait a bit before trying again
-        # final trial, die on exception
-            self.retry_count = n+1
-            return fn(self, *args)
-        except Exception, e:
-            log.Debug(_("Backtrace of previous error: %s")
-                        % exception_traceback())
-            log.FatalError(_("Giving up after %s attempts. %s: %s")
-                         % (self.retry_count, e.__class__.__name__, util.uexc(e)),
-                          log.ErrorCode.backend_error)
-        self.retry_count = 0
+                    at_end = n == globals.num_retries
+                    code = _get_code_from_exception(self.backend, operation, e)
+                    if code == log.ErrorCode.backend_not_found:
+                        # If we tried to do something, but the file just isn't there,
+                        # no need to retry.
+                        at_end = True
+                    if at_end and fatal:
+                        def make_filename(f):
+                            if isinstance(f, path.ROPath):
+                                return util.escape(f.name)
+                            else:
+                                return util.escape(f)
+                        extra = ' '.join([operation] + [make_filename(x) for x in args if x])
+                        log.FatalError(_("Giving up after %s attempts. %s: %s")
+                                       % (n, e.__class__.__name__,
+                                          util.uexc(e)), code=code, extra=extra)
+                    else:
+                        log.Warn(_("Attempt %s failed. %s: %s")
+                                 % (n, e.__class__.__name__, util.uexc(e)))
+                    if not at_end:
+                        if isinstance(e, TemporaryLoadException):
+                            time.sleep(3 * globals.backend_retry_delay)  # wait longer before trying again
+                        else:
+                            time.sleep(globals.backend_retry_delay)  # wait a bit before trying again
+                        if hasattr(self.backend, '_retry_cleanup'):
+                            self.backend._retry_cleanup()
+
+        return inner_retry
+    return outer_retry
 
-    return _retry_fatal
 
 class Backend(object):
     """
-    Represents a generic duplicity backend, capable of storing and
-    retrieving files.
-
-    Concrete sub-classes are expected to implement:
-
-      - put
-      - get
-      - list
-      - delete
-      - close (if needed)
-
-    Optional:
-
-      - move
+    See README in backends directory for information on how to write a backend.
     """
-    
     def __init__(self, parsed_url):
         self.parsed_url = parsed_url
 
-    def put(self, source_path, remote_filename = None):
-        """
-        Transfer source_path (Path object) to remote_filename (string)
-
-        If remote_filename is None, get the filename from the last
-        path component of pathname.
-        """
-        raise NotImplementedError()
-
-    def move(self, source_path, remote_filename = None):
-        """
-        Move source_path (Path object) to remote_filename (string)
-
-        Same as put(), but unlinks source_path in the process.  This allows the
-        local backend to do this more efficiently using rename.
-        """
-        self.put(source_path, remote_filename)
-        source_path.delete()
-
-    def get(self, remote_filename, local_path):
-        """Retrieve remote_filename and place in local_path"""
-        raise NotImplementedError()
-
-    def list(self):
-        """
-        Return list of filenames (byte strings) present in backend
-        """
-        def tobytes(filename):
-            "Convert a (maybe unicode) filename to bytes"
-            if isinstance(filename, unicode):
-                # There shouldn't be any encoding errors for files we care
-                # about, since duplicity filenames are ascii.  But user files
-                # may be in the same directory.  So just replace characters.
-                return filename.encode(sys.getfilesystemencoding(), 'replace')
-            else:
-                return filename
-
-        if hasattr(self, '_list'):
-            # Make sure that duplicity internals only ever see byte strings
-            # for filenames, no matter what the backend thinks it is talking.
-            return map(tobytes, self._list())
-        else:
-            raise NotImplementedError()
-
-    def delete(self, filename_list):
-        """
-        Delete each filename in filename_list, in order if possible.
-        """
-        raise NotImplementedError()
-
-    # Should never cause FatalError.
-    # Returns a dictionary of dictionaries.  The outer dictionary maps
-    # filenames to metadata dictionaries.  Supported metadata are:
-    #
-    # 'size': if >= 0, size of file
-    #         if -1, file is not found
-    #         if None, error querying file
-    #
-    # Returned dictionary is guaranteed to contain a metadata dictionary for
-    # each filename, but not all metadata are guaranteed to be present.
-    def query_info(self, filename_list, raise_errors=True):
-        """
-        Return metadata about each filename in filename_list
-        """
-        info = {}
-        if hasattr(self, '_query_list_info'):
-            info = self._query_list_info(filename_list)
-        elif hasattr(self, '_query_file_info'):
-            for filename in filename_list:
-                info[filename] = self._query_file_info(filename)
-
-        # Fill out any missing entries (may happen if backend has no support
-        # or its query_list support is lazy)
-        for filename in filename_list:
-            if filename not in info:
-                info[filename] = {}
-
-        return info
-
     """ use getpass by default, inherited backends may overwrite this behaviour """
     use_getpass = True
 
@@ -483,7 +430,7 @@ class Backend(object):
         except KeyError:
             if self.use_getpass:
                 password = getpass.getpass("Password for '%s@%s': " %
-                                           (self.parsed_url.username,self.parsed_url.hostname) )
+                                           (self.parsed_url.username, self.parsed_url.hostname))
                 os.environ['FTP_PASSWORD'] = password
             else:
                 password = None
@@ -499,42 +446,28 @@ class Backend(object):
         the ':password@' may be substituted.
         """
         if self.parsed_url.password:
-            return re.sub( r'(:([^\s:/@]+)@([^\s@]+))', r':*****@\3', commandline )
+            return re.sub(r'(:([^\s:/@]+)@([^\s@]+))', r':*****@\3', commandline)
         else:
             return commandline
 
-    """
-    DEPRECATED:
-    run_command(_persist) - legacy wrappers for subprocess_popen(_persist)
-    """
-    def run_command(self, commandline):
-        return self.subprocess_popen(commandline)
-    def run_command_persist(self, commandline):
-        return self.subprocess_popen_persist(commandline)
-
-    """
-    DEPRECATED:
-    popen(_persist) - legacy wrappers for subprocess_popen(_persist)
-    """
-    def popen(self, commandline):
-        result, stdout, stderr = self.subprocess_popen(commandline)
-        return stdout
-    def popen_persist(self, commandline):
-        result, stdout, stderr = self.subprocess_popen_persist(commandline)
-        return stdout
-
-    def _subprocess_popen(self, commandline):
+    def __subprocess_popen(self, args):
         """
         For internal use.
         Execute the given command line, interpreted as a shell command.
         Returns int Exitcode, string StdOut, string StdErr
         """
         from subprocess import Popen, PIPE
-        p = Popen(commandline, shell=True, stdout=PIPE, stderr=PIPE)
+
+        args[0] = util.which(args[0])
+        p = Popen(args, stdout=PIPE, stderr=PIPE)
         stdout, stderr = p.communicate()
 
         return p.returncode, stdout, stderr
 
+    """ a dictionary for breaking exceptions, syntax is
+        { 'command' : [ code1, code2 ], ... } see ftpbackend for an example """
+    popen_breaks = {}
+
     def subprocess_popen(self, commandline):
         """
         Execute the given command line with error check.
@@ -542,58 +475,194 @@ class Backend(object):
 
         Raise a BackendException on failure.
         """
-        private = self.munge_password(commandline)
-        log.Info(_("Reading results of '%s'") % private)
-        result, stdout, stderr = self._subprocess_popen(commandline)
+        import shlex
+
+        if isinstance(commandline, (types.ListType, types.TupleType)):
+            logstr = ' '.join(commandline)
+            args = commandline
+        else:
+            logstr = commandline
+            args = shlex.split(commandline)
+
+        logstr = self.munge_password(logstr)
+        log.Info(_("Reading results of '%s'") % logstr)
+
+        result, stdout, stderr = self.__subprocess_popen(args)
         if result != 0:
-            raise BackendException("Error running '%s'" % private)
-        return result, stdout, stderr
-
-    """ a dictionary for persist breaking exceptions, syntax is
-        { 'command' : [ code1, code2 ], ... } see ftpbackend for an example """
-    popen_persist_breaks = {}
-
-    def subprocess_popen_persist(self, commandline):
-        """
-        Execute the given command line with error check.
-        Retries globals.num_retries times with 30s delay.
-        Returns int Exitcode, string StdOut, string StdErr
-
-        Raise a BackendException on failure.
-        """
-        private = self.munge_password(commandline)
-
-        for n in range(1, globals.num_retries+1):
-            # sleep before retry
-            if n > 1:
-                time.sleep(30)
-            log.Info(_("Reading results of '%s'") % private)
-            result, stdout, stderr = self._subprocess_popen(commandline)
-            if result == 0:
-                return result, stdout, stderr
-
             try:
-                m = re.search("^\s*([\S]+)", commandline)
-                cmd = m.group(1)
-                ignores = self.popen_persist_breaks[ cmd ]
+                ignores = self.popen_breaks[args[0]]
                 ignores.index(result)
                 """ ignore a predefined set of error codes """
                 return 0, '', ''
             except (KeyError, ValueError):
-                pass
+                raise BackendException("Error running '%s': returned %d, with output:\n%s" %
+                                       (logstr, result, stdout + '\n' + stderr))
+        return result, stdout, stderr
 
-            log.Warn(ngettext("Running '%s' failed with code %d (attempt #%d)",
-                              "Running '%s' failed with code %d (attempt #%d)", n) %
-                               (private, result, n))
-            if stdout or stderr:
-                    log.Warn(_("Error is:\n%s") % stderr + (stderr and stdout and "\n") + stdout)
 
-        log.Warn(ngettext("Giving up trying to execute '%s' after %d attempt",
-                          "Giving up trying to execute '%s' after %d attempts",
-                          globals.num_retries) % (private, globals.num_retries))
-        raise BackendException("Error running '%s'" % private)
+class BackendWrapper(object):
+    """
+    Represents a generic duplicity backend, capable of storing and
+    retrieving files.
+    """
 
-    def get_fileobj_read(self, filename, parseresults = None):
+    def __init__(self, backend):
+        self.backend = backend
+
+    def __do_put(self, source_path, remote_filename):
+        if hasattr(self.backend, '_put'):
+            log.Info(_("Writing %s") % util.ufn(remote_filename))
+            self.backend._put(source_path, remote_filename)
+        else:
+            raise NotImplementedError()
+
+    @retry('put', fatal=True)
+    def put(self, source_path, remote_filename=None):
+        """
+        Transfer source_path (Path object) to remote_filename (string)
+
+        If remote_filename is None, get the filename from the last
+        path component of pathname.
+        """
+        if not remote_filename:
+            remote_filename = source_path.get_filename()
+        self.__do_put(source_path, remote_filename)
+
+    @retry('move', fatal=True)
+    def move(self, source_path, remote_filename=None):
+        """
+        Move source_path (Path object) to remote_filename (string)
+
+        Same as put(), but unlinks source_path in the process.  This allows the
+        local backend to do this more efficiently using rename.
+        """
+        if not remote_filename:
+            remote_filename = source_path.get_filename()
+        if hasattr(self.backend, '_move'):
+            if self.backend._move(source_path, remote_filename) is not False:
+                source_path.setdata()
+                return
+        self.__do_put(source_path, remote_filename)
+        source_path.delete()
+
+    @retry('get', fatal=True)
+    def get(self, remote_filename, local_path):
+        """Retrieve remote_filename and place in local_path"""
+        if hasattr(self.backend, '_get'):
+            self.backend._get(remote_filename, local_path)
+            local_path.setdata()
+            if not local_path.exists():
+                raise BackendException(_("File %s not found locally after get "
+                                         "from backend") % util.ufn(local_path.name))
+        else:
+            raise NotImplementedError()
+
+    @retry('list', fatal=True)
+    def list(self):
+        """
+        Return list of filenames (byte strings) present in backend
+        """
+        def tobytes(filename):
+            "Convert a (maybe unicode) filename to bytes"
+            if isinstance(filename, unicode):
+                # There shouldn't be any encoding errors for files we care
+                # about, since duplicity filenames are ascii.  But user files
+                # may be in the same directory.  So just replace characters.
+                return filename.encode(globals.fsencoding, 'replace')
+            else:
+                return filename
+
+        if hasattr(self.backend, '_list'):
+            # Make sure that duplicity internals only ever see byte strings
+            # for filenames, no matter what the backend thinks it is talking.
+            return [tobytes(x) for x in self.backend._list()]
+        else:
+            raise NotImplementedError()
+
+    def delete(self, filename_list):
+        """
+        Delete each filename in filename_list, in order if possible.
+        """
+        assert not isinstance(filename_list, types.StringType)
+        if hasattr(self.backend, '_delete_list'):
+            self._do_delete_list(filename_list)
+        elif hasattr(self.backend, '_delete'):
+            for filename in filename_list:
+                self._do_delete(filename)
+        else:
+            raise NotImplementedError()
+
+    @retry('delete', fatal=False)
+    def _do_delete_list(self, filename_list):
+        while filename_list:
+            sublist = filename_list[:100]
+            self.backend._delete_list(sublist)
+            filename_list = filename_list[100:]
+
+    @retry('delete', fatal=False)
+    def _do_delete(self, filename):
+        self.backend._delete(filename)
+
+    # Should never cause FatalError.
+    # Returns a dictionary of dictionaries.  The outer dictionary maps
+    # filenames to metadata dictionaries.  Supported metadata are:
+    #
+    # 'size': if >= 0, size of file
+    #         if -1, file is not found
+    #         if None, error querying file
+    #
+    # Returned dictionary is guaranteed to contain a metadata dictionary for
+    # each filename, and all metadata are guaranteed to be present.
+    def query_info(self, filename_list):
+        """
+        Return metadata about each filename in filename_list
+        """
+        info = {}
+        if hasattr(self.backend, '_query_list'):
+            info = self._do_query_list(filename_list)
+            if info is None:
+                info = {}
+        elif hasattr(self.backend, '_query'):
+            for filename in filename_list:
+                info[filename] = self._do_query(filename)
+
+        # Fill out any missing entries (may happen if backend has no support
+        # or its query_list support is lazy)
+        for filename in filename_list:
+            if filename not in info or info[filename] is None:
+                info[filename] = {}
+            for metadata in ['size']:
+                info[filename].setdefault(metadata, None)
+
+        return info
+
+    @retry('query', fatal=False)
+    def _do_query_list(self, filename_list):
+        info = self.backend._query_list(filename_list)
+        if info is None:
+            info = {}
+        return info
+
+    @retry('query', fatal=False)
+    def _do_query(self, filename):
+        try:
+            return self.backend._query(filename)
+        except Exception as e:
+            code = _get_code_from_exception(self.backend, 'query', e)
+            if code == log.ErrorCode.backend_not_found:
+                return {'size': -1}
+            else:
+                raise e
+
+    def close(self):
+        """
+        Close the backend, releasing any resources held and
+        invalidating any file objects obtained from the backend.
+        """
+        if hasattr(self.backend, '_close'):
+            self.backend._close()
+
+    def get_fileobj_read(self, filename, parseresults=None):
         """
         Return fileobject opened for reading of filename on backend
 
@@ -608,38 +677,7 @@ class Backend(object):
         tdp.setdata()
         return tdp.filtered_open_with_delete("rb")
 
-    def get_fileobj_write(self, filename,
-                          parseresults = None,
-                          sizelist = None):
-        """
-        Return fileobj opened for writing, which will cause the file
-        to be written to the backend on close().
-
-        The file will be encoded as specified in parseresults (or as
-        read from the filename), and stored in a temp file until it
-        can be copied over and deleted.
-
-        If sizelist is not None, it should be set to an empty list.
-        The number of bytes will be inserted into the list.
-        """
-        if not parseresults:
-            parseresults = file_naming.parse(filename)
-            assert parseresults, u"Filename %s not correctly parsed" % util.ufn(filename)
-        tdp = dup_temp.new_tempduppath(parseresults)
-
-        def close_file_hook():
-            """This is called when returned fileobj is closed"""
-            self.put(tdp, filename)
-            if sizelist is not None:
-                tdp.setdata()
-                sizelist.append(tdp.getsize())
-            tdp.delete()
-
-        fh = dup_temp.FileobjHooked(tdp.filtered_open("wb"))
-        fh.addhook(close_file_hook)
-        return fh
-
-    def get_data(self, filename, parseresults = None):
+    def get_data(self, filename, parseresults=None):
         """
         Retrieve a file from backend, process it, return contents.
         """
@@ -647,18 +685,3 @@ class Backend(object):
         buf = fin.read()
         assert not fin.close()
         return buf
-
-    def put_data(self, buffer, filename, parseresults = None):
-        """
-        Put buffer into filename on backend after processing.
-        """
-        fout = self.get_fileobj_write(filename, parseresults)
-        fout.write(buffer)
-        assert not fout.close()
-
-    def close(self):
-        """
-        Close the backend, releasing any resources held and
-        invalidating any file objects obtained from the backend.
-        """
-        pass
