@@ -491,6 +491,11 @@ def parse_cmdline_options(arglist):
                       callback=lambda o, s, v, p: (setattr(p.values, o.dest, True),
                                                    old_fn_deprecation(s)))
 
+    # Sync only required metadata
+    parser.add_option("--metadata-sync-mode",
+                      default="full",
+                      choices=("full", "partial"))
+
     # Level of Redundancy in % for Par2 files
     parser.add_option("--par2-redundancy", type="int", metavar=_("number"))
 
@@ -623,11 +628,6 @@ def parse_cmdline_options(arglist):
     parser.add_option("--volsize", type="int", action="callback", metavar=_("number"),
                       callback=lambda o, s, v, p: setattr(p.values, "volsize", v * 1024 * 1024))
 
-    # If set, collect only the file status, not the whole root.
-    parser.add_option("--file-changed", action="callback", type="file",
-                      metavar=_("path"), dest="file_changed",
-                      callback=lambda o, s, v, p: setattr(p.values, "file_changed", v.rstrip('/')))
-
     # delay time before next try after a failure of a backend operation
     # TRANSL: Used in usage help. Example:
     # --backend-retry-delay <seconds>
